#
# Module to provide general utility functions for Astrolabe code.
#   Written by: Tom Hicks. 7/26/2018.
#   Last Modified: Remove unused, leftover code.
#
import fnmatch
import os
import pathlib as pl

# Default text file of desired metadata keys, one per line
_DEFAULT_KEYS_FILE = "metadata-keys.txt"

# patterns for identifying FITS and gzipped FITS files
_FITS_PAT = "*.fits"
_GZFITS_PAT = "*.fits.gz"

def is_fits_file(fyl):
    """ Return True if the given file is FITS file, else False. """
    return (fnmatch.fnmatch(fyl, _FITS_PAT) or fnmatch.fnmatch(fyl, _GZFITS_PAT))

def filter_file_tree(root_dir):
    """ Generator to yield all FITS files in the file tree under the given root directory. """
    for root, dirs, files in os.walk(root_dir):
        for fyl in files:
            if (is_fits_file(fyl)):
                file_path = os.path.join(root, fyl)
                yield file_path

def get_metadata_keys(options):
    """ Return a list of metadata keys to be extracted. """
    keyfile = options.get("keyfile")
    if (keyfile):
        with open(keyfile, "r") as mdkeys_file:
            return mdkeys_file.read().splitlines()
    else:
        return None

def path_has_dots(apath):
    """ Tell whether the given path contains '.' or '..' """
    parts = list(pl.PurePath(apath).parts)
    return ((apath == ".") or (".." in parts) or ("." in parts))



#
# Module to view, extract, and/or verify metadata from one or more FITS files.
#   Written by: Tom Hicks. 4/24/2018.
#   Last Modified: Sequence the special cases. Revert to using right_ascension & declination names.
#
import os
import sys
import warnings
from astropy.io import fits
import astrolabe_py.utils as utils
from astrolabe_py.fits_meta import FitsMeta

# dictionary of alternates for standard FITS metadata keys
_ALTERNATE_KEYS_MAP = {
    "right_ascension": "s_ra",              # calculated -> VO
    "declination": "s_dec",                 # calculated -> VO
    "DATE-BEG": "t_min",                    # FITS -> VO
    "DATE-END": "t_max",                    # FITS -> VO
    "DATE-OBS": "t_min",                    # FITS -> VO
    "INSTRUME": "instrument_name",          # FITS -> VO
    "MJD-END" : "t_max",                    # FITS -> VO
    "MJD-OBS" : "t_min",                    # FITS -> VO
    "NAXIS1"  : "sxel1",                    # FITS -> VO
    "NAXIS2"  : "sxel2",                    # FITS -> VO
    "OBJECT"  : "target_name",              # FITS -> VO
    "OBSERVER": "obs_creator_name",         # FITS -> VO
    "TELESCOP": "facility_name",            # FITS -> VO
    "TSTART"  : "t_min",                    # FITS -> VO
    "TSTOP"   : "t_max"                     # FITS -> VO
}

# dictionary mapping CTYPE* key names to their associated CRVAL* key names
_CTYPES = { "CTYPE1": "CRVAL1",  "CTYPE2": "CRVAL2" }


def execute_info(options):
    """ Returns a (possibly empty) list, each element of which is a list of
        summary information strings for the HDUs in a single FITS file.
    """
    # execute action for a single file or a directory of files
    file_path = options.get("images_path")
    if (os.path.isfile(file_path)):
        info = fits_hdu_info(file_path, options)
        return [info] if info else []
    else:
        if (os.path.isdir(file_path)):
            info_lst = [fits_hdu_info(fits_file, options)
                        for fits_file in utils.filter_file_tree(file_path)]
            return [info for info in info_lst if info] # just in case: remove empty lists
        else:                                          # should never happen
            print("Error: Specified file path '{}' is not a file or directory".format(file_path))
            sys.exit(20)

def fits_hdu_info(file_path, options={}):
    """ Return a list of summary information strings for the HDUs of the given FITS file. """
    verbose = options.get("verbose", False)
    if (verbose):
        print("Reading HDU information for file {} ...".format(file_path))
    fm = FitsMeta(file_path)
    hduinfo = fm.hdu_info()
    filename = os.path.basename(fm.filepath())
    # format the information into a report (a list of strings):
    results = ["Filename: {}".format(filename),
               "No.    Name      Ver    Type      Cards   Dimensions   Format"]
    layout = "{:3d}  {:10}  {:3} {:11}  {:5d}   {}   {}   {}"
    for hinfo in hduinfo:
        results.append(layout.format(*hinfo))
    return results


def fits_metadata(file_path, options={}):
    """ Return a list Metadatum tuples extracted from the given FITS file. """
    keys_subset = options.get("keys_subset")
    ignore_keys = options.get("ignore_keys")
    fm = FitsMeta(file_path, ignore_keys=ignore_keys)
    metadata = _post_process_metadata(fm, keys_subset)
    return metadata


def _post_process_metadata(fm, keys_subset):
    """ Post process the accumulated metadata; handle a couple of special cases. """
    for item in fm.metadata():              # check all metadata items for CTYPE special cases
        _handle_ctype_mapping(fm, item, keys_subset) # fm and key_subset modified by side-effect

    for item in fm.metadata():              # check all metadata items for alternate keys
        _handle_alternate_key(fm, item, keys_subset) # fm and key_subset modified by side-effect

    if (keys_subset):                       # if user requested only a subset of the metadata
        return fm.filter_by_keys(keys_subset) # filter the metadata by the keys subset
    else:
        return fm.metadata()                # else just return all the accumulated metadata


def _handle_alternate_key(fm, item, keys_subset):
    """ For items whose keys are listed in the alternate key table, duplicate the item
        but use the alternate keyword for the duplicated item. """
    item_key = item.keyword                 # keyword of this item
    alt_key = _ALTERNATE_KEYS_MAP.get(item_key) # try to get an alternate key for this item
    if (alt_key):                           # if an alternate key exists
        if (keys_subset):                   # if only a subset of keys requested
            if (item_key in keys_subset):   # and this item key is in that subset
                if (fm.copy_item(item_key, alt_key)): # copy standard item w/ alternate keyword
                    keys_subset.append(alt_key) # if copied, add alternate keyword to the subset
                else:                           # else copy failed: move on
                    pass
            else:                           # else key not in subset: ignore this item
                pass
        else:                               # else not using a subset, so copy item
            fm.copy_item(item_key, alt_key) # copy standard item w/ alternate keyword


def _handle_ctype_mapping(fm, item, keys_subset):
    """ If a metadata item has a CTYPE key, it holds the interpretation of a corresponding
        CRVAL metadata item. Add a new item with the 'interpretation' key and the CRVAL value.
        For CRVALs and how they relate to CTYPEs see https://fits.gsfc.nasa.gov/fits_standard.html
    """
    # if this item's key is a CTYPE key, then get the CRVAL key interpreted by this item:
    crval_key = _CTYPES.get(item.keyword)   # lookup this item's key in CTYPE dictionary
    if (crval_key):                         # if this item key is a CTYPE key
        if "RA" in item.value:              # if this CTYPE item's value contains RA
            interp_key = "right_ascension"  # the 'interpretation' of the CRVAL value
        elif "DEC" in item.value:           # else if this CTYPE item's value contains DEC
            interp_key = "declination"      # the 'interpretation' of the CRVAL value
        else:                               # we only handle these interpretations, so far
            interp_key = None

        if (interp_key):                    # if we have a workable intepretation
            copied = fm.copy_item(crval_key, interp_key) # copy CRVAL value w/ interpretation key
            if (copied and keys_subset):       # if only a subset of keys requested
                keys_subset.append(interp_key) # add the interpretation keyword to the subset


def execute_verify(options):
    """ Returns a (possibly empty) list, each element of which is a list of
        warning strings from files which violate the FITS standard.
    """
    # execute action for a single file or a directory of files
    file_path = options.get("images_path")
    if (os.path.isfile(file_path)):
        warn = fits_verify(file_path, options)
        return [warn] if warn else []
    else:
        if (os.path.isdir(file_path)):
            warn_lst = [fits_verify(fits_file, options)
                        for fits_file in utils.filter_file_tree(file_path)]
            return [warn for warn in warn_lst if warn] # remove empty lists
        else:                                          # should never happen
            print("Error: Specified file path '{}' is not a file or directory".format(file_path))
            sys.exit(30)

def fits_verify(file_path, options={}):
    """ Verify that the data in the given FITS file conforms to the FITS standard.
        Return a (possibly empty) list of verification warning strings.
    """
    verbose = options.get("verbose", False)
    results = []
    with fits.open(file_path) as hdu:
        if (verbose):
            print("Checking file {} ...".format(file_path))
        with warnings.catch_warnings(record=True) as warns:
            hdu.verify("fix+warn")
            if (warns and len(warns) > 0):
                results.append("Filename: {}".format(file_path))
                for warn in warns:
                    results.append(str(warn.message))
    return results



#!/usr/bin/env python
#
# Setup script.
#   Written by: Tom Hicks. 6/22/2018.
#   Last Modified: Figure out requires. Redo VERSION reading.
#
import os
import re
import sys
from setuptools import setup, find_packages

# Eval the VERSION number from the version file
with open('astrolabe_py/version.py') as vfile:
    exec(vfile.read())

# Read in the long description from the README file
with open("README.rst", "r", encoding="UTF-8") as rm:
    long_description = rm.read()

# Build the list of scripts to be installed.
# script_dir = 'scripts'
# scripts = []
# for script in os.listdir(script_dir):
#     scripts.append(os.path.join(script_dir,script))

setup(
    name="astrolabe_py",
    version=VERSION,
    author="Tom Hicks",
    author_email="hickst@email.arizona.edu",
    description="Tools for working with Astrolabe data.",
    license='Apache Software License (http://www.apache.org/licenses/LICENSE-2.0)',
    long_description=long_description,
    long_description_content_type="text/x-rst",
    url="https://github.com/AstrolabeProject/astrolabe_py",
    packages=find_packages(exclude=['test.*']),
    package_data ={
        'astrolabe_py': ['data/*']
    },
    install_requires=[
        'astropy>=3.0.3',
        'python-irodsclient>=0.8.1'
    ],
    python_requires='~=3.6',
    # scripts=scripts,
    scripts=[ "checker", "uploader" ],
    classifiers=[
        'Development Status :: 1 - Planning',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: End Users/Desktop',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: POSIX',
        'Programming Language :: Python :: 3',
        'Topic :: System :: Archiving',
    ],
    keywords='Astrolabe datastore iRods metadata uploading'
)



#
# Module to extract metadata and upload one or more FITS files to iRods.
#   Written by: Tom Hicks. 7/19/2018.
#   Last Modified: Maintain return of same truth vector as previous version.
#
import os
import sys
import logging
import fnmatch

import astrolabe_py.fits_ops as fo
import astrolabe_py.irods_help as ih
import astrolabe_py.utils as utils

logging.basicConfig(level=logging.INFO)     # default logging configuration

_ASTROLABE_ROOT_DIR = "astrolabe"

def execute(options):
    """ Process the specified file(s) according to the other given arguments.
        This module operates on one or more FITS files to extract metadata,
        uploads the files to iRods, and attach the metadata to the files.
    """
    # create connection to iRods
    ihelper = ih.IrodsHelper()
    ensure_astrolabe_root(ihelper)          # create/use astrolabe directory, as needed

    # get the desired subset of metadata keys, if any specified by a keyfile
    md_keys = utils.get_metadata_keys(options)
    if (md_keys):
        options["keys_subset"] = md_keys

    # execute action for a single file or a directory of files
    images_path = options.get("images_path")
    if (utils.path_has_dots(images_path)):
        print("Error: Images path argument must be asbolute; it may not contain '..' or '.'")
        sys.exit(10)

    if (os.path.isfile(images_path)):
        to_path = options.get("to_path")    # allow for future expansion
        if (not to_path):
            to_path = os.path.basename(images_path)
        return [ do_file(ihelper, images_path, to_path, options) ]
    else:
        if (os.path.isdir(images_path)):
            return do_tree(ihelper, images_path, options)
        else:
            print("Error: Specified images path '{}' is not a file or directory".format(images_path))
            sys.exit(11)


def do_file(ihelper, source_file, to_path, options):
    """ Do metadata extraction, file upload, and metadata attachment for the given file,
        depending on the settings of the various arguments in the given 'options' dictionary.
    """
    verbose = options.get("verbose", False)
    upload_only = options.get("upload_only", False)

    if (verbose):
        print("Uploading file {} to {}".format(source_file, to_path))
    ihelper.put_file(source_file, to_path, absolute=True)

    if ((not upload_only) and has_metadata(source_file)):
        if (verbose):
            print("Extracting metadata from file {}".format(source_file))
        metadata = fo.fits_metadata(source_file, options)
        if (verbose):
            print("Attaching metadata to file {}".format(to_path))
        ihelper.put_metaf(metadata, to_path)

    return True


def do_tree(ihelper, root_node, options):
    """ Walk the local filesystem tree from the given root_node and process any FITS files.
        The walk creates a parallel tree in the iRods Astrolabe area and calls do_file to
        upload the files (and possibly their metadata) to the corresponding iRods directories.
    """
    root_path = os.path.normpath(root_node) # remove any trailing slashes from given root path

    # get a list of files to be uploaded
    source_paths = get_source_paths(root_path, options)

    # remove the root path prefix from the upload files
    suffix_paths = make_suffix_paths(root_path, source_paths, options)

    # make a list of directories to be created in iRods and create them
    dir_paths = make_dir_paths(suffix_paths, options)
    for adir in dir_paths:
        ihelper.mkdir(adir, absolute=True)  # make corresponding iRods directory

    # make a list of user-home-relative target file paths
    target_paths = make_target_paths(suffix_paths, options)

    # pair up the local source file paths and the iRods target file paths, then upload the files
    return [ do_file(ihelper, pair[0], pair[1], options) for pair in zip(source_paths, target_paths) ]


def ensure_astrolabe_root(ihelper):
    """ Ensure that the Astrolabe root directory exists and set it as the user's root directory. """
    ihelper.set_root()                      # reset root to user iRod root dir
    ihelper.cd_root()                       # set cwd to user iRod root dir
    ihelper.mkdir(_ASTROLABE_ROOT_DIR)      # make the Astrolabe directory
    ihelper.set_root(top_dir=_ASTROLABE_ROOT_DIR) # reset root to Astrolabe dir


def get_source_paths(root_path, options):
    """Walk the given root path, returning a list of paths for files to be uploaded."""
    results = []
    for root, dirs, files in os.walk(root_path):
        for afile in files:                 # for files at this level
            if (isa_desired_file(afile)):   # if a file is to be uploaded
                results.append(os.path.join(root, afile)) # local source path
    return results


def has_metadata(afile):
    """Return true if the given file is one that contains metadata. Currently, only FITS files."""
    return (utils.is_fits_file(afile))      # if a file is a FITS file

def isa_desired_file(afile):
    """Return true if the given file is one that is to be uploaded. Currently, only FITS files."""
    return (utils.is_fits_file(afile))      # if a file is a FITS file


def make_dir_paths(file_paths, options):
    """Return a sorted list of unique directory paths, extracted from the given list
       of file paths.
    """
    return sorted({os.path.split(path)[0] for path in file_paths}, reverse=True)


def make_suffix_paths(root_path, source_paths, options):
    """Remove the given root path from each of the paths in the given source path list."""
    source_prefix = os.path.split(root_path)[0] # find local path prefix preceding root node
    # print("PREFIX: {}".format(source_prefix))
    return [os.path.relpath(path, source_prefix) for path in source_paths]


def make_target_paths(suffix_paths, options, irods_prefix=_ASTROLABE_ROOT_DIR):
    """Return a list of user-home-relative iRods paths for the given list of local file paths."""
    return suffix_paths                     # currently a NOP
    # return [os.path.join(irods_prefix, afile) for afile in suffix_paths]
